/*
  Copyright (c) 2003, Joerg Raedler, <jr@j-raedler.de>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
  Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
  Neither the name of the author nor the names of its contributors may
  be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#include <string.h>
#include "Python.h"
#include "structmember.h"

#include "agrepy_all.h"

#ifndef FUGREPVERSION
#define FUGREPVERSION 0.0
#endif

#ifndef FUGREPCVSID
#define FUGREPCVSID "$Id: cFuGrep.c,v 1.10 2003-11-20 14:27:35 joerg Exp $"
#endif

#define FUGREP_DOC "\nFuGrep means 'fuzzy grep'. It's based on agrep "\
"and uses parts and concepts of the older agrep-python-interface called agrepy.\n\n"\
"The main part of this module is a type called 'Pattern'. "\
"Use an instance of 'Pattern' to search a pattern with some fault tolerance "\
"in a text string.\n"\
"An example:\n"\
">>> from FuGrep import Pattern\n"\
">>> pat = Pattern('fuzy')\n"\
">>> print pat.grep('FuGrep means fuzzy grep!')\n"\
"[(13, 17)]\n"\
"There's one match in the string between index 13 and 17 matching the word "\
"'fuzy' with one difference.\n\n"\
"A function called similarity(s0, s1) may be used to compare two strings. "\
"It returns 1.0 for equality, 0.0 for complete difference and a value "\
"between 0.0 and 1.0 for the level of similarity:\n"\
">>> from FuGrep import similarity\n"\
">>> similarity('wether', 'weather')\n"\
"0.92307692307692313\n"\
">>> similarity('foo', 'bar')\n"\
"0.0\n\n"

#define PATTERN_DOC "\n  Pattern(pattern, nErrors=1) - Fuzzy Pattern object.\n"\
"'pattern' is a pattern string, 'nErrors' the allowed number of errors for a match.\n\n"

#define GREP_DOC "\n  grep(text, greedy=True) - grep a string\n"\
"text is the string to search the pattern in. If greedy is unset, "\
"the matches are smaller. This function returns a list of 2-tuples containing the "\
"start and and index of a match in the string.\n\n"

#define SIMILARITY_DOC "\n  similarity(s0, s1) - compare two strings\n"\
"compare s0 and s1, return a float value between 0.0 and 1.0. A value of 1.0 "\
"means equality, zero indicates no similarity between the two strings."\
"This function may be used to sort a list of spelling alternatives to a given word.\n\n"


/* collect results from the grep routines into a python list */
int_pair_list *add_ends(int start, int end, int_pair_list *matches) {
    PyObject *t;
    if (! matches) matches = PyList_New(0);
    t = PyTuple_New(2);
    PyTuple_SetItem(t, 0, PyInt_FromLong(start));
    PyTuple_SetItem(t, 1, PyInt_FromLong(end+1));
    PyList_Append(matches, t);
    return matches;
}


/* recursive implementation of the Ratcliff/Obershelp Pattern
Recognition Algorithm as described in the July 1988 issue of Dr. Dobbs
Journal. This was written using parts of the concept of a file simil.c
implemented by Steve Grubb <linux_4ever@yahoo.com> */
static void str_similarity_rec(double *v, char *s0, char *s1, int l0, int l1) {
    int a = 0, b = -1, len, i, j, size = 1;
    for (i=0; i<=(l0-size); i++) {
        for (j=0; j<=(l1-size); j++) {
            len = size;
            while (42) {
                if ((len<=(l0-i))&&(len<=(l1-j))) {
                    if (memcmp(s0+i, s1+j, len) == 0) {
                        if ((len > size)||(b < 0)) {
                            a = i; b = j; size = len;
                        }
                        len++;
                    }
                    else break;
                }
                else break;
            }
        }
    }
    if ((len = ((b<0)?0:size))) {
        *v += len;
        str_similarity_rec(v, s0, s1, a, b);
        i = a+len;
        j = b+len;
        if ((i<l0)&&(j<l1))
            str_similarity_rec(v, s0+i, s1+j, l0-i, l1-j);
    }
}


static PyObject *Str_Similarity(PyObject *self, PyObject *args) {
    char *s0, *s1;
    int l0, l1;
    double v = 0.0;
    if (!PyArg_ParseTuple(args, "ss", &s0, &s1))
        return NULL;
    if ((l0 = strlen(s0)) && (l1 = strlen(s1))) {
        str_similarity_rec(&v, s0, s1, l0, l1);
        return Py_BuildValue("d", 2.0*v/(l0+l1));
    } else
        return Py_BuildValue("d", 0.0);
}


typedef struct {
    PyObject_HEAD
    int nErr;
    int plen;
    char *pat;
    param_struct *param;
} Pattern;


static void Pattern_dealloc(Pattern* self) {
    /*if (self->text)  free(self->text);*/
    if (self->param) free(self->param);
    self->ob_type->tp_free((PyObject*)self);
}


static PyObject *Pattern_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
    Pattern *self = (Pattern *)type->tp_alloc(type, 0);
    if (self != NULL) {
        self->nErr = 1;
        self->plen = 0;
        self->pat = NULL;
        self->param = NULL;
    }
    return (PyObject *)self;
}


static int Pattern_init(Pattern *self, PyObject *args, PyObject *kwds) {
    static char *kwlist[] = {"pattern", "nErrors", NULL};
    self->nErr = 1;
    if (! PyArg_ParseTupleAndKeywords(args, kwds, "s|i", kwlist, &(self->pat), &(self->nErr)))
        return -1;
    self->plen = strlen(self->pat);
    if (self->plen > MAXPATLEN) {
        PyErr_SetString(PyExc_Exception, "Pattern is to long!");
        return -1;
    }
    if ((self->nErr < 1) || (self->nErr > 8)) {
        PyErr_SetString(PyExc_Exception, "Number of errors must be between 1 and 8!");
        return -1;
    }
    if (self->nErr >= self->plen) {
        PyErr_SetString(PyExc_Exception, "Number of errors must not exceed the pattern size!");
        return -1;
    }
    if (self->plen <= SHORT_LONG)
        self->param = (param_struct*)sagrepy_compile(self->pat, self->plen, self->nErr);
    else
        self->param = (param_struct*)lagrepy_compile(self->pat, self->plen, self->nErr);
    return 0;
}


static PyObject *Pattern_grep(Pattern* self, PyObject *args) {
    char *text;
    int greedy=1;
    PyObject *res, *tmp=NULL;
    if (! PyArg_ParseTuple(args, "s|O", &text, &tmp))
        return NULL;
    if (tmp)
        greedy = PyObject_IsTrue(tmp);
    if (self->plen <= SHORT_LONG)
        res = exec_sagrepy(self->pat, self->plen, text, strlen(text), greedy, self->param);
    else
        res = exec_lagrepy(self->pat, self->plen, text, strlen(text), greedy, self->param);
    if (! res) res = Py_None;
    Py_INCREF(res);
    return res;
}


static PyObject *Pattern_repr(Pattern *self) {
    return PyString_FromFormat("<FuGrep.Pattern '%s' %d>",
                               self->pat, self->nErr);
}


static int Pattern_print(Pattern *self, FILE *fp, int flags) {
    fprintf(fp, "<FuGrep.pattern '%s', %d>", self->pat, self->nErr);
    return 0;
}

static PyMemberDef Pattern_members[] = {
    {"pattern", T_STRING, offsetof(Pattern, pat), READONLY, "pattern (readonly)"},
    {"nErrors", T_INT, offsetof(Pattern, nErr), READONLY, "number of allowed errors (readonly)"},
    {NULL}  /* Sentinel */
};


static PyMethodDef Pattern_methods[] = {
    {"grep", (PyCFunction)Pattern_grep, METH_VARARGS, GREP_DOC},
    {NULL}  /* Sentinel */
};


static PyTypeObject PatternType = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "cFuGrep.Pattern",          /*tp_name*/
    sizeof(Pattern),           /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    (destructor)Pattern_dealloc, /*tp_dealloc*/
    (printfunc)Pattern_print,  /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    (reprfunc)Pattern_repr,    /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    PyObject_GenericGetAttr,   /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    PATTERN_DOC,               /* tp_doc */
    0,                         /* tp_traverse */
    0,                         /* tp_clear */
    0,                         /* tp_richcompare */
    0,                         /* tp_weaklistoffset */
    0,                         /* tp_iter */
    0,                         /* tp_iternext */
    Pattern_methods,           /* tp_methods */
    Pattern_members,           /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)Pattern_init,    /* tp_init */
    0,                         /* tp_alloc */
    Pattern_new,                /* tp_new */
};


static PyMethodDef FuGrep_methods[] = {
    {"similarity", (PyCFunction)Str_Similarity, METH_VARARGS, SIMILARITY_DOC},
    {NULL}  /* Sentinel */
};


#ifndef PyMODINIT_FUNC  /* declarations for DLL import/export */
#define PyMODINIT_FUNC void
#endif
PyMODINIT_FUNC initcFuGrep(void) {
    PyObject *m, *d;
    PatternType.tp_new = PyType_GenericNew;
    if (PyType_Ready(&PatternType) < 0)
        return;
    m = Py_InitModule3("cFuGrep", FuGrep_methods, FUGREP_DOC);
    Py_INCREF(&PatternType);
    PyModule_AddObject(m, "Pattern", (PyObject *)&PatternType);

    d = PyModule_GetDict(m);
    PyDict_SetItemString(d, "version", (PyObject *)PyFloat_FromDouble(FUGREPVERSION));
    PyDict_SetItemString(d, "_cvsId", (PyObject *)PyString_FromString(FUGREPCVSID));
    PyDict_SetItemString(d, "_author", (PyObject *)PyString_FromString("Joerg Raedler | dezentral | software@dezentral.de"));
    PyDict_SetItemString(d, "_homepage", (PyObject *)PyString_FromString("http://www.dezentral.de/soft/FuGrep"));
}
